from dataclasses import dataclass
import torch
import logging
import abc
from arm_pytorch_utilities import tensor_utils, serialization

logger = logging.getLogger(__name__)


@dataclass
class ContactParameters:
    length: float = 0.1
    penetration_length: float = 0.01
    hard_assignment_threshold: float = 0.4  # for soft assignment, probability threshold for belonging to same component
    intersection_tolerance: float = 0.002  # how much intersection into the robot's surface we ignore


# resample function from http://scipy-cookbook.readthedocs.io/items/ParticleFilter.html
def resample(weights):
    n = len(weights)
    indices = []
    C = [0.0] + [torch.sum(weights[: i + 1]) for i in range(n)]
    u0, j = torch.rand((1,), device=weights.device), 0
    for u in [(u0 + i) / n for i in range(n)]:
        while j < len(C) and u > C[j]:
            j += 1
        indices.append(j - 1)
    return indices


class ContactSet(serialization.Serializable):
    def __init__(self, params: ContactParameters, immovable_collision_checker=None, device='cpu', dtype=torch.float32,
                 visualizer=None, dim=2):
        self.device = device
        self.dtype = dtype
        self.p = params
        self.immovable_collision_checker = immovable_collision_checker
        self.visualizer = visualizer
        self.dim = dim

    @abc.abstractmethod
    def update(self, x, dx, p=None, info=None):
        """Update contact set with observed transition

        :param x: current end effector pose
        :param dx: latest change in end effector pose while in contact
        :param p: current position of the contact point, or none if not in contact
        :param info: dictionary with optional keys from InfoKeys that provide extra debugging information
        """

    @abc.abstractmethod
    def get_batch_data_for_dynamics(self, total_num):
        """Return initial contact data needed for dynamics for total_num batch size"""

    @abc.abstractmethod
    def dynamics(self, x, u, contact_data):
        """Perform batch dynamics on x with control u using contact data from get_batch_data_for_dynamics

        :return next x, without contact mask, updated contact data"""


class LinearTranslationalDynamics:
    @staticmethod
    def _extract_linear_translation(dx):
        # for base, x is just position so dx is linear translation
        return dx

    def __call__(self, p, x, dx):
        dpos = LinearTranslationalDynamics._extract_linear_translation(dx)
        new_p = p + dpos
        new_x = x + dx
        return new_p, new_x


class ContactSetSoft(ContactSet):
    """
    Track contact points and contact configurations without hard object assignments (partitions of points that move
    together). Implemented using a particle filter with each particle being the set of all contact points and configs.
    """

    def __init__(self, pxpen, *args, pxdyn=LinearTranslationalDynamics(), n_particles=100, n_eff_threshold=0.8,
                 replace_bad_points=True, **kwargs):
        super(ContactSetSoft, self).__init__(*args, **kwargs)
        # key functions
        self.pxpen = pxpen
        self.pxdyn = pxdyn

        self.adjacency = None
        self.connection_prob = None
        self.n_particles = n_particles
        self.n_eff_threshold = n_eff_threshold
        self._do_replace_bad_points = replace_bad_points

        self.weights = torch.ones(self.n_particles, device=self.device) / self.n_particles
        self.weights_normalization = 1

        # each observation has a contact point (on robot surface) and contact config (pose of end effector)
        self.pts = None
        self.configs = None
        self.acts = None

        # [n_particles, N, D] where N is the number of data points, and D is data dimensionality
        self.sampled_pts = None
        self.sampled_configs = None
        self.map_particle = None

    def _compute_full_adjacency(self, pts):
        # don't typically need to compute full adjacency
        dd = torch.cdist(pts, pts)
        return self._distance_to_probability(dd)

    def get_posterior_points(self):
        return self.map_particle

    def _partition_points(self, adjacent):
        # holds a partition of indices 0..N-1
        groups = []
        N = adjacent.shape[0]
        all_selected = torch.zeros(N, device=adjacent.device, dtype=torch.bool)
        i = 0
        while not all_selected.all():
            if all_selected[i]:
                i += 1
                continue
            this_group = adjacent[i].clone()
            # TODO should use some connected component finding algo; for simplicity for now just look at 1 depth
            for j in range(N):
                if this_group[j]:
                    this_group |= adjacent[j]
            groups.append(this_group)
            all_selected |= this_group
        return groups

    def get_hard_assignment(self, threshold=None):
        pts = self.get_posterior_points()
        if pts is None:
            return []
        connection_prob = self._compute_full_adjacency(pts)

        if threshold is None:
            threshold = torch.rand_like(connection_prob)

        adjacent = threshold < connection_prob
        return self._partition_points(adjacent)

    def get_partitioned_points(self, threshold):
        pts = self.get_posterior_points()
        partitions = self.get_hard_assignment(threshold)
        return [pts[partition] for partition in partitions]

    def _distance_to_probability(self, distance, sigma=None):
        # parameter where higher means a greater drop off in probability with distance
        if sigma is None:
            sigma = 1 / self.p.length
        return torch.exp(-sigma * distance ** 2)

    def predict_particles(self, query_pts, dx):
        """Apply action to all particles"""
        # dd = (self.pts[-1] - self.sampled_pts).norm(dim=-1)
        dd = torch.cdist(query_pts, self.sampled_pts)

        # each point can at most be affected by 1 dx per contact
        # we compare its sampled probability against the most likely motion
        # take the minimum distance to any of our query points for probability
        dd, closest_dx_index = dd.min(dim=-2)

        # convert to probability
        connection_prob = self._distance_to_probability(dd)

        # sample particles which make hard assignments
        # independently sample uniform [0, 1) and compare against prob - note that connections are symmetric
        # sampled_prob[i,j] is the ith particle's probability of connection between the latest point and the jth point
        sampled_prob = torch.rand(connection_prob.shape, device=self.pts.device)
        adjacent = sampled_prob < connection_prob

        # don't actually need to label connected components because we just need to propagate for the latest
        # apply dx to each particle's cluster that contains the latest x
        self.sampled_pts[adjacent], self.sampled_configs[adjacent] = self.pxdyn(self.sampled_pts[adjacent],
                                                                                self.sampled_configs[adjacent],
                                                                                dx[closest_dx_index][adjacent])
        return adjacent

    def update_particles(self, config=None):
        """Update the weight of each particle corresponding to their ability to explain the observation"""
        if self.pts is None:
            return

        # all contact points should be outside the robot
        tol = self.p.intersection_tolerance
        # if given an explicit config to check against
        if config is not None:
            # for efficiency, just consider the given configuration (should probably consider all points, but may be enough)
            query_points = self.sampled_pts.view(-1, self.sampled_pts.shape[-1])
            d = self.pxpen(config.view(1, -1), query_points).view(self.n_particles, -1)
            # negative distance indicates penetration
            d += tol
            d[d > 0] = 0
            # collect sum then offset by max to prevent obs_weights from going to 0
            obs_weights = d.sum(dim=1)
        else:
            # otherwise check all configs against all points for each particle
            obs_weights = torch.zeros(self.n_particles, dtype=self.pts.dtype, device=self.pts.device)
            for i in range(self.n_particles):
                d = self.pxpen(self.sampled_configs[i], self.sampled_pts[i])
                d += tol
                d = d[d < 0]
                obs_weights[i] = d.sum()

        # prevent every particle going to 0
        obs_weights -= obs_weights.max()
        # convert to probability
        penetration_sigma = 1 / self.p.penetration_length
        obs_weights = self._distance_to_probability(obs_weights, sigma=penetration_sigma)

        min_weight = 1e-15
        self.weights = self.weights * obs_weights
        self.weights[self.weights < min_weight] = min_weight

        # normalize weights to resampling probabilities
        self.weights_normalization = self.weights.sum()
        self.weights = self.weights / self.weights_normalization

        # from pfilter
        # Compute effective sample size and entropy of weighting vector.
        # These are useful statistics for adaptive particle filtering.
        self.n_eff = (1.0 / (self.weights ** 2).sum()) / self.n_particles
        self.weight_entropy = torch.sum(self.weights * torch.log(self.weights))
        logger.debug(f"PF total weights: {self.weights_normalization.item()} n_eff {self.n_eff.item()}")

        # preserve current sample set before any replenishment
        self.original_particles = self.sampled_pts.clone(), self.sampled_configs.clone()

        # store MAP estimate
        argmax_weight = torch.argmax(self.weights)
        self.map_particle = self.sampled_pts[argmax_weight]

        # resampling (systematic resampling) step
        if self.n_eff < self.n_eff_threshold:
            indices = resample(self.weights)
            self.sampled_pts = self.sampled_pts[indices, :]
            self.sampled_configs = self.sampled_configs[indices, :]

            # if even after resampling we have bad points, replace those points with good points
            resampled_weights = self.weights[indices]
            resampled_n_eff = (1.0 / (resampled_weights ** 2).sum()) / self.n_particles
            if resampled_n_eff < self.n_eff_threshold and self._do_replace_bad_points:
                self.replace_bad_points()

            self.weights = torch.ones(self.n_particles, device=self.device) / self.n_particles

    def replace_bad_points(self):
        # for points that remain low probability, replace them with a non penetrating one
        for i in range(self.n_particles):
            d = self.pxpen(self.sampled_configs[i], self.sampled_pts[i])
            d += self.p.intersection_tolerance
            d[d > 0] = 0
            # each column is the distance of that point to each config
            bad_pts = d.sum(dim=0) < 0
            if not torch.any(bad_pts):
                continue
            good_pts = ~bad_pts
            # can't do anything if we have no good points...
            if not torch.any(good_pts):
                continue
            # find the distance of bad points wrt good points
            d_to_good = torch.cdist(self.sampled_pts[i, bad_pts], self.sampled_pts[i, good_pts])
            # replace the bad point with a copy of the closest good point
            closest_idx = torch.argmin(d_to_good, dim=1)
            self.sampled_pts[i, bad_pts] = self.sampled_pts[i, good_pts][closest_idx]
            self.sampled_configs[i, bad_pts] = self.sampled_configs[i, good_pts][closest_idx]

    def update(self, x, dx, p=None, info=None):
        d = self.device
        dtype = self.dtype
        x = tensor_utils.ensure_tensor(d, dtype, x)

        if info is not None:
            u = info['u']
        else:
            u = torch.zeros_like(x)

        cur_config = x
        if p is None or dx is None:
            # step without contact, eliminate particles that conflict with this config in freespace
            self.update_particles(cur_config)
            return None, None

        dx = tensor_utils.ensure_tensor(d, dtype, dx)
        # handle multiple number of points
        # where contact point would be without this movement
        if len(p.shape) < 2:
            p = p.reshape(1, -1)
        N = len(p)
        cur_pt = p[:, :self.dim]
        # 1 dx per sensor
        prev_pt, prev_config = self.pxdyn(cur_pt.view(N, -1), cur_config.view(1, -1), -dx)
        u = u.repeat(N, 1)
        # both should be (N, -1)

        if self.pts is None:
            self.pts = prev_pt
            self.configs = prev_config
            self.acts = u
            self.sampled_pts = self.pts.repeat(self.n_particles, 1, 1)
            self.sampled_configs = self.configs.repeat(self.n_particles, 1, 1)
        else:
            self.pts = torch.cat((self.pts, prev_pt))
            self.configs = torch.cat((self.configs, prev_config))
            self.acts = torch.cat((self.acts, u))
            self.sampled_pts = torch.cat([self.sampled_pts, prev_pt.repeat(self.n_particles, 1, 1)], dim=1)
            self.sampled_configs = torch.cat([self.sampled_configs, prev_config.repeat(self.n_particles, 1, 1)], dim=1)

        # classic alternation of predict and update steps
        self.predict_particles(prev_pt, dx)
        # check all configs against all points
        self.update_particles(None)

        return True, True

    def dynamics(self, x, u, contact_data):
        raise NotImplementedError()

    def get_batch_data_for_dynamics(self, total_num):
        raise NotImplementedError()



import torch
import math
import abc
import typing
from collections import deque

import pytorch_kinematics.transforms as tf
from arm_pytorch_utilities import math_utils, tensor_utils
import numpy as np

point = torch.tensor
points = torch.tensor
normals = torch.tensor


class ContactSensor:
    """An independent sensor that holds sensor related information during an action"""

    def __init__(self, dtype=torch.float, device='cpu'):
        self.in_contact = False
        self.dx = []
        self.dtype = dtype
        self.device = device

    def clear(self):
        self.in_contact = False
        self.dx = []

    def observe_residual(self, residual):
        pass

    def get_dx(self):
        dx = np.sum(self.dx, axis=0)
        return torch.tensor(dx, dtype=self.dtype, device=self.device)

    def observe_dx(self, dx):
        if self.in_contact:
            self.dx.append(dx)

    @abc.abstractmethod
    def isolate_contact(self, ee_force_torque, pose, q=None, visualizer=None):
        """Return contact point in link frame that most likely explains the observed residual"""


class ResidualPlanarContactSensor(ContactSensor):
    """Contact sensor based on generalized momentum residual measurements only"""

    def __init__(self, surface_points, surface_normals, residual_threshold, max_friction_cone_angle=50 * math.pi / 180,
                 dtype=torch.float, device='cpu'):
        """
        :param surface_points: N x 3 sampled points on the robot surface in link frame
        :param surface_normals: N x 3 corresponding surface normals to the surface points, only necessary for
        visualization
        :param residual_threshold: contact threshold for residual^T sigma_meas^-1 residual to count as being in contact
        :param max_friction_cone_angle: max angular difference (radian) from surface normal that a contact is possible
        note that 90 degrees is the weakest assumption that the force is from a push and not a pull
        """
        self._cached_points = surface_points
        self._cached_normals = surface_normals

        self._residual_threshold = residual_threshold
        self.max_friction_cone_angle = max_friction_cone_angle

        super().__init__(dtype=dtype, device=device)

    def observe_residual(self, residual):
        self.in_contact = residual > self._residual_threshold

    def get_jacobian(self, locations, q=None):
        """Get J^T in the equation: wrench at end effector = J^T * wrench at contact point.
        In general the Jacobian is dependent on configuration.

        Locations are specified wrt the end effector frame.
        For planar robots, this kind of Jacobian is configuration independent"""

        return torch.stack(
            [torch.tensor([[1., 0.], [0., 1.], [-loc[1], loc[0]]], device=self.device, dtype=self.dtype) for loc in
             locations])

    def isolate_contact(self, ee_force_torque, pose, q=None, visualizer=None):
        # 2D
        link_frame_pts, pts, normals = self.sample_robot_surface_points(pose, visualizer=visualizer)
        F_c = ee_force_torque[:, :2]
        T_ee = ee_force_torque[:, -1]

        while True:
            # reject points where force is sufficiently away from surface normal
            from_normal = math_utils.angle_between(normals[:, :2], -F_c)
            valid = from_normal < self.max_friction_cone_angle
            # validity has to hold across all experienced forces
            valid = valid.all(dim=1)
            # remove a single element of the constraint if we don't have any satisfying them
            if valid.any():
                break
            else:
                F_c = F_c[:-1]
                T_ee = T_ee[:-1]

        # no valid point
        if F_c.numel() == 0:
            return None

        pts = pts[valid]
        link_frame_pts = link_frame_pts[valid]

        # NOTE: if our system is able to rotate, would have to transform points by rotation too
        # get relative to end effector origin
        rel_pts = pts - pose[0]
        J = self.get_jacobian(rel_pts, q=q)
        # J_{r_c}^T F_c
        expected_residual = J @ F_c.transpose(-1, -2)

        # the below is the case for full residual; however we can shortcut since we only need to compare torque
        # error = ee_force_torque - expected_residual
        # combined_error = linalg.batch_quadratic_product(error, self.residual_precision)

        error = expected_residual[:, -1] - T_ee
        # don't have to worry about normalization since it's just the torque dimension
        combined_error = error.abs().sum(dim=1)

        min_err_i = torch.argmin(combined_error)

        if visualizer is not None:
            # also draw some other likely points
            likely_pt_index = torch.argsort(combined_error)
            for i in reversed(range(1, min(10, len(pts)))):
                pt = pts[likely_pt_index[i]]
                visualizer.draw_point(f'likely.{i}', pt, height=pt[2] + 0.001,
                                      color=(0, 1 - 0.8 * i / 10, 0))
            visualizer.draw_point(f'most likely contact', pts[min_err_i], color=(0, 1, 0), scale=2)
            visualizer.draw_2d_line('reaction', pts[min_err_i], ee_force_torque.mean(dim=0)[:3], color=(0, 0.2, 1.0),
                                    scale=0.2)

        return link_frame_pts[min_err_i]

    def sample_robot_surface_points(self, pose, visualizer=None) -> typing.Tuple[points, points, normals]:
        """Get points on the surface of the robot that could be possible contact locations
        pose[0] and pose[1] are the position and orientation (quaternion) of the end effector, respectively.
        Also return the correspnoding surface normals for each of the points.
        """
        if self._cached_points.dtype != self.dtype or self._cached_points.device != self.device:
            self._cached_points = self._cached_points.to(device=self.device, dtype=self.dtype)
            self._cached_normals = self._cached_normals.to(device=self.device, dtype=self.dtype)

        link_to_current_tf = tf.Transform3d(pos=pose[0], rot=tf.xyzw_to_wxyz(
            tensor_utils.ensure_tensor(self.device, self.dtype, pose[1])), dtype=self.dtype, device=self.device)
        pts = link_to_current_tf.transform_points(self._cached_points)
        normals = link_to_current_tf.transform_normals(self._cached_normals)
        if visualizer is not None:
            for i, pt in enumerate(pts):
                visualizer.draw_point(f't.{i}', pt, color=(1, 0, 0), height=pt[2])
                visualizer.draw_2d_line(f'n.{i}', pt, normals[i], color=(0.5, 0, 0), size=2., scale=0.1)

        return self._cached_points, pts, normals


class ContactDetector:
    """Detect and isolate contacts given some form of generalized momentum residual measurements, see
    https://ieeexplore.ieee.org/document/7759743 (Localizing external contact using proprioceptive sensors).

    We additionally assume access to force torque sensors at the end effector, which is our residual."""

    def __init__(self, residual_precision, window_size=5, dtype=torch.float, device='cpu'):
        """
        :param residual_precision: sigma_meas^-1 matrix that scales the different residual dimensions based on their
        expected precision
        """

        self.residual_precision = residual_precision
        self.observation_history = deque(maxlen=500)
        self._window_size = window_size
        self.dtype = dtype
        self.device = device

        self.sensors: typing.List[ContactSensor] = []

    def register_contact_sensor(self, sensor: ContactSensor):
        sensor.dtype = self.dtype
        sensor.device = self.device
        self.sensors.append(sensor)

    def to(self, device=None, dtype=None):
        if device is not None:
            self.device = device
            for sensor in self.sensors:
                sensor.device = device
        if dtype is not None:
            self.dtype = dtype
            for sensor in self.sensors:
                sensor.dtype = dtype

    def observe_residual(self, ee_force_torque, pose=None):
        """Returns whether this residual implies we are currently in contact and track its location if given pose"""
        epsilon = ee_force_torque.T @ self.residual_precision @ ee_force_torque

        in_contact = False

        for sensor in self.sensors:
            sensor.observe_residual(epsilon)
            if sensor.in_contact:
                in_contact = True

        self.observation_history.append((in_contact, ee_force_torque, pose))

        return in_contact

    def observe_dx(self, dx):
        for sensor in self.sensors:
            sensor.observe_dx(dx)

    def clear_sensors(self):
        for sensor in self.sensors:
            sensor.clear()

    def __len__(self):
        return len(self.observation_history)

    def clear(self):
        self.observation_history.clear()
        self.clear_sensors()

    def isolate_contact(self, ee_force_torque, pose, q=None, visualizer=None):
        """Return contact points in link frame that most likely explains the observed residual,
        and the change in config associated with that contact"""
        pts = torch.stack(
            [sensor.isolate_contact(ee_force_torque, pose, q=q, visualizer=visualizer) for sensor in self.sensors])
        dx = torch.stack([sensor.get_dx() for sensor in self.sensors])
        return pts, dx

    def in_contact(self):
        """Whether our last observed residual indicates that we are currently in contact"""
        if len(self.observation_history) == 0:
            return False
        for i in range(self._window_size):
            index = -1 - i
            # out of range
            if index < -len(self.observation_history):
                return False
            in_contact, ee_force_torque, prev_pose = self.observation_history[index]
            if in_contact:
                return True
        return False

    def get_last_contact_location(self, pose=None, **kwargs):
        """Get last contact point given the current end effector pose"""
        if len(self.observation_history) == 0:
            return None, None

        # allow for being in contact anytime within the latest window
        start_i = -1
        for i in range(0, min(len(self.observation_history), self._window_size)):
            in_contact, ee_force_torque, prev_pose = self.observation_history[-1 - i]
            if in_contact:
                if pose is None:
                    pose = prev_pose
                break
            start_i -= 1
        else:
            return None, None

        # use history of points to handle jitter
        ee_ft = [ee_force_torque]
        pp = [prev_pose]
        for i in range(2, self._window_size):
            from_end_i = start_i - i + 1
            if -from_end_i > len(self.observation_history):
                break
            in_contact, ee_force_torque, prev_pose = self.observation_history[from_end_i]
            # look back until we leave contact
            if not in_contact:
                break
            ee_ft.append(ee_force_torque)
            pp.append(prev_pose)
        ee_ft = torch.tensor(np.stack(ee_ft), dtype=self.dtype, device=self.device)
        # assume we don't move that much in a short amount of time and we can just use the latest pose
        pp = tuple(torch.tensor(p, dtype=self.dtype, device=self.device) for p in pp[0])
        # pos = torch.tensor([p[0] for p in pp], dtype=self.dtype)
        # orientation = torch.tensor([p[1] for p in pp], dtype=self.dtype)
        # pp = (pos, orientation)

        # in link frame
        last_contact_point, dx = self.isolate_contact(ee_ft, pp, **kwargs)
        if last_contact_point is None:
            return None, None

        link_to_current_tf = tf.Transform3d(pos=pose[0], rot=tf.xyzw_to_wxyz(torch.tensor(pose[1])),
                                            dtype=self.dtype, device=self.device)
        return link_to_current_tf.transform_points(last_contact_point), dx



import torch
from arm_pytorch_utilities import tensor_utils


class PlanarMovableSDF:
    """Measure distance from a point to the robot surface for the robot in some configuration"""

    def __init__(self, d_cache, min_x, min_y, max_x, max_y, cache_resolution, cache_y_len):
        self.d_cache = d_cache
        self.min_x, self.min_y, self.max_x, self.max_y = min_x, min_y, max_x, max_y
        self.cache_y_len = cache_y_len
        self.cache_resolution = cache_resolution
        # for debugging, view a = self.d_cache.reshape(len(x), len(y))

    def __call__(self, configs, pts):
        if not torch.is_tensor(self.d_cache):
            self.d_cache = tensor_utils.ensure_tensor(pts.device, pts.dtype, self.d_cache)

        M = configs.shape[0]
        N = pts.shape[0]
        # take advantage of the fact our system has no rotation to just translate points; otherwise would need full tsf
        query_pts = pts.repeat(M, 1, 1).transpose(0, 1)
        # needed transpose to allow broadcasting
        query_pts -= configs
        # flatten
        query_pts = query_pts.transpose(0, 1).view(M * N, -1)
        d = torch.ones(M * N, dtype=pts.dtype, device=pts.device)
        # eliminate points outside bounding box
        valid = (self.min_x <= query_pts[:, 0]) & (query_pts[:, 0] <= self.max_x) & \
                (self.min_y <= query_pts[:, 1]) & (query_pts[:, 1] <= self.max_y)
        # convert coordinates to indices
        query = query_pts[valid]
        x_id = torch.round((query[:, 0] - self.min_x) / self.cache_resolution).to(dtype=torch.long)
        y_id = torch.round((query[:, 1] - self.min_y) / self.cache_resolution).to(dtype=torch.long)
        idx = x_id * self.cache_y_len + y_id
        d[valid] = self.d_cache[idx]

        return d.view(M, N)





import torch
from stucco.tracking import ContactSetSoft, ContactParameters
from arm_pytorch_utilities.rand import seed
import matplotlib.pyplot as plt
from matplotlib.patches import Circle


class CircularPointToConfig:
    """Measure distance from a point to the robot surface for the robot in some configuration"""

    def __init__(self, robot_radius):
        self.r = robot_radius

    def __call__(self, configs, pts):
        M = configs.shape[0]
        N = pts.shape[0]

        # take advantage of the fact our system has no rotation to just translate points; otherwise would need full tsf
        query_pts = pts.repeat(M, 1, 1).transpose(0, 1)
        # needed transpose to allow broadcasting
        query_pts -= configs
        # flatten
        query_pts = query_pts.transpose(0, 1).view(M * N, -1)

        d = query_pts.norm(dim=-1) - self.r

        return d.view(M, N)


PREV_PT_RGBA = (0.1, 0.8, 0.1, 0.5)
CUR_PT_RGBA = "#FFA500"


def plot_and_update(r, contact_set, x, dx, p):
    seed(1)

    fig, ax = plt.subplots()
    ax.axis('equal')
    ax.grid(True, linestyle='--')
    ax.set(xlim=(-1.5, 1.5), ylim=(-1.5, 1.5))
    s = 12

    # plot robot
    prev_x = x - dx
    prev_robot = Circle(prev_x[0], r, color=(0.8, 0.1, 0.1, 0.5), zorder=0)
    ax.add_patch(prev_robot)

    prev_pt = p - dx

    robot = Circle(x, r, color=(0.8, 0.1, 0.1, 0.8), zorder=1)
    ax.add_patch(robot)

    prev_sampled_pts = None
    if contact_set.sampled_pts is not None:
        prev_sampled_pts = contact_set.sampled_pts.clone()

    contact_set.update(x, dx, p)

    for i in range(contact_set.n_particles):
        ax.set_title(f'particle {i}')
        actors = []
        cx = contact_set.sampled_pts[i, :, 0].view(1, -1)
        cy = contact_set.sampled_pts[i, :, 1].view(1, -1)

        lx = prev_pt[:, 0].view(-1)
        ly = prev_pt[:, 1].view(-1)
        if prev_sampled_pts is not None:
            px = torch.cat((prev_sampled_pts[i, :, 0], lx))
            py = torch.cat((prev_sampled_pts[i, :, 1], ly))
        else:
            px = lx
            py = ly

        actors.append(ax.scatter(px, py, s=s, c=[PREV_PT_RGBA], zorder=2))
        actors.append(ax.quiver(px, py, cx - px, cy - py, scale=1, width=0.01, units='xy', zorder=3))
        actors.append(ax.scatter(cx, cy, s=s, c=[CUR_PT_RGBA for _ in cx], zorder=4))
        actors.append(ax.scatter(p[:, 0], p[:, 1], s=s, c=[(0, 0, 0)], zorder=5))

        plt.pause(0.05)
        plt.savefig(f'/home/zhsh/Pictures/fig.png')

        user_input = input('enter for next, q for break\n')

        if user_input == 'q':
            break

        for actor in actors:
            actor.remove()


def test_multiple_dynamics():
    r = 0.1
    pt_to_config = CircularPointToConfig(r)
    params = ContactParameters(length=0.5, penetration_length=0.1, hard_assignment_threshold=0.4,
                               intersection_tolerance=0.0)

    contact_set = ContactSetSoft(pt_to_config, params, n_particles=10)

    # simulation
    x = torch.tensor([0., 0.])

    plt.ion()
    plt.show()

    x = x + torch.tensor([0.3, 0.])
    dx = torch.tensor([[0.3, 0.], [0.15, 0.]])
    p = torch.tensor([[0.4, 0.23], [0.4, 0.4]])
    plot_and_update(r, contact_set, x, dx, p)

    x = x + torch.tensor([0.2, 0.])
    dx = torch.tensor([[0.2, 0.], [0.1, 0.]])
    p = torch.tensor([[0.6, -0.15], [0.6, -0.25]])
    plot_and_update(r, contact_set, x, dx, p)

    x = x + torch.tensor([0.3, 0.])
    dx = torch.tensor([[0.3, 0.], [0.15, 0.]])
    p = torch.tensor([[0.9, -0.35], [0.9, 0.55]])
    plot_and_update(r, contact_set, x, dx, p)

def test_single_dynamics():
    r = 0.5
    pt_to_config = CircularPointToConfig(r)
    params = ContactParameters(length=0.5, penetration_length=0.1, hard_assignment_threshold=0.4,
                               intersection_tolerance=0.0)

    contact_set = ContactSetSoft(pt_to_config, params, n_particles=10)

    # simulation
    x = torch.tensor([0., 0.])
    dx = torch.tensor([0.3, 0.])

    plt.ion()
    plt.show()

    x = x + dx
    plot_and_update(r, contact_set, x, dx.view(1, -1), torch.tensor([[0.4, 0.23]]))

    dx = torch.tensor([0.1, 0.])
    x = x + dx
    plot_and_update(r, contact_set, x, dx.view(1, -1), torch.tensor([[0.5, 0.15]]))

    dx = torch.tensor([0.1, 0.3])
    x = x + dx
    plot_and_update(r, contact_set, x, dx.view(1, -1), torch.tensor([[0.5, 0.]]))

    dx = torch.tensor([0.1, -0.3])
    x = x + dx
    plot_and_update(r, contact_set, x, dx.view(1, -1), torch.tensor([[0.6, -0.1]]))


if __name__ == "__main__":
    # test_single_dynamics()
    test_multiple_dynamics()
